Package: rprojroot
Title: Finding Files in Project Subdirectories
Version: 2.0.4
Authors@R: 
    person(given = "Kirill",
           family = "M\u00fcller",
           role = c("aut", "cre"),
           email = "kirill@cynkra.com",
           comment = c(ORCID = "0000-0002-1416-3412"))
Description: Robust, reliable and flexible paths to files below
    a project root. The 'root' of a project is defined as a directory that
    matches a certain criterion, e.g., it contains a certain regular file.
License: MIT + file LICENSE
URL: https://rprojroot.r-lib.org/, https://github.com/r-lib/rprojroot
BugReports: https://github.com/r-lib/rprojroot/issues
Depends: R (>= 3.0.0)
Suggests: covr, knitr, lifecycle, mockr, rlang, rmarkdown, testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: no
Packaged: 2023-11-05 06:47:23 UTC; kirill
Author: Kirill Müller [aut, cre] (<https://orcid.org/0000-0002-1416-3412>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2023-11-05 10:20:02 UTC
Built: R 4.4.0; ; 2026-07-17 17:58:44 UTC; unix

Package: rprojroot
Title: Finding Files in Project Subdirectories
Version: 2.1.1
Authors@R: 
    person(given = "Kirill",
           family = "M\u00fcller",
           role = c("aut", "cre"),
           email = "kirill@cynkra.com",
           comment = c(ORCID = "0000-0002-1416-3412"))
Description: Robust, reliable and flexible paths to files below
    a project root. The 'root' of a project is defined as a directory that
    matches a certain criterion, e.g., it contains a certain regular file.
License: MIT + file LICENSE
URL: https://rprojroot.r-lib.org/, https://github.com/r-lib/rprojroot
BugReports: https://github.com/r-lib/rprojroot/issues
Depends: R (>= 3.0.0)
Suggests: covr, knitr, lifecycle, rlang, rmarkdown, testthat (>=
        3.2.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.2.9000
Config/autostyle/scope: line_breaks
Config/autostyle/strict: true
Config/Needs/website: tidyverse/tidytemplate
NeedsCompilation: no
Packaged: 2025-08-26 15:22:42 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2025-08-26 16:20:02 UTC
Built: R 4.6.0; ; 2026-04-26 09:05:43 UTC; unix

Package: rprojroot
Title: Finding Files in Project Subdirectories
Version: 2.0.3
Authors@R: 
    person(given = "Kirill",
           family = "M\u00fcller",
           role = c("aut", "cre"),
           email = "krlmlr+r@mailbox.org",
           comment = c(ORCID = "0000-0002-1416-3412"))
Description: Robust, reliable and flexible paths to files below
    a project root. The 'root' of a project is defined as a directory that
    matches a certain criterion, e.g., it contains a certain regular file.
License: MIT + file LICENSE
URL: https://rprojroot.r-lib.org/, https://github.com/r-lib/rprojroot
BugReports: https://github.com/r-lib/rprojroot/issues
Depends: R (>= 3.0.0)
Suggests: covr, knitr, lifecycle, mockr, rmarkdown, testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.1.2
NeedsCompilation: no
Packaged: 2022-04-02 16:14:00 UTC; kirill
Author: Kirill Müller [aut, cre] (<https://orcid.org/0000-0002-1416-3412>)
Maintainer: Kirill Müller <krlmlr+r@mailbox.org>
Repository: CRAN
Date/Publication: 2022-04-02 16:40:02 UTC

Package: hierarchy
Title: What the Package Does (one line, title case)
Version: 0.0-0
Authors@R: as.person("Kirill Müller <krlmlr+r@mailbox.org> [aut, cre]")
Description: What the package* does.
Depends: R (>= 3.2.0)
License: GPL-3
LazyData: true
Encoding: UTF-8

# Generated by roxygen2: do not edit by hand

S3method("|",root_criterion)
S3method(as_root_criterion,character)
S3method(as_root_criterion,default)
S3method(as_root_criterion,root_criterion)
S3method(format,root_criterion)
S3method(print,root_criterion)
S3method(str,root_criteria)
export(as.root_criterion)
export(as_root_criterion)
export(criteria)
export(find_package_root_file)
export(find_remake_root_file)
export(find_root)
export(find_root_file)
export(find_rstudio_root_file)
export(find_testthat_root_file)
export(from_wd)
export(get_root_desc)
export(has_basename)
export(has_dir)
export(has_file)
export(has_file_pattern)
export(is.root_criterion)
export(is_drake_project)
export(is_git_root)
export(is_pkgdown_project)
export(is_projectile_project)
export(is_quarto_project)
export(is_r_package)
export(is_remake_project)
export(is_renv_project)
export(is_root_criterion)
export(is_rstudio_project)
export(is_svn_root)
export(is_testthat)
export(is_vcs_root)
export(root_criterion)
export(thisfile)
export(thisfile_knit)
export(thisfile_r)
export(thisfile_rscript)
export(thisfile_source)
importFrom(utils,str)
importFrom(utils,tail)

<!-- README.md and index.md are generated from README.Rmd. Please edit that file. -->

# [rprojroot](https://rprojroot.r-lib.org/)

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![rcc](https://github.com/r-lib/rprojroot/workflows/rcc/badge.svg)](https://github.com/r-lib/rprojroot/actions)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/rprojroot)](https://cran.r-project.org/package=rprojroot)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/rprojroot/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/rprojroot?branch=main)
<!-- badges: end -->

This package helps accessing files relative to a *project root* to [stop
the working directory
insanity](https://gist.github.com/jennybc/362f52446fe1ebc4c49f). It is a
low-level helper package for the [here](https://here.r-lib.org/)
package.

``` r
library(rprojroot)
```

## Example

The rprojroot package works best when you have a “project”: all related
files contained in a subdirectory that can be categorized using a strict
criterion. Let’s create a package for demonstration.

``` r
dir <- tempfile()
pkg <- usethis::create_package(dir)
#> ✔ Creating
#>   /var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586/.
#> ✔ Setting active project to
#>   "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586".
#> ✔ Creating R/.
#> ✔ Writing DESCRIPTION.
#> Package: file337b29a46586
#> Title: What the Package Does (One Line, Title Case)
#> Version: 0.0.0.9000
#> Date: 2025-06-27
#> Authors@R (parsed):
#>     * Kirill Müller <kirill@cynkra.com> [aut, cre] (<https://orcid.org/0000-0002-1416-3412>)
#> Description: What the package does (one paragraph).
#> License: MIT
#> URL: https://github.com/krlmlr/rprojroot,
#>     https://krlmlr.github.io/rprojroot
#> BugReports: https://github.com/krlmlr/rprojroot/issues
#> Encoding: UTF-8
#> Roxygen: list(markdown = TRUE)
#> RoxygenNote: 7.3.2.9000
#> ✔ Writing NAMESPACE.
#> ✔ Setting active project to "<no active project>".
```

R packages satisfy the `is_r_package` criterion. A criterion is an
object that contains a `find_file()` function. With `pkg` as working
directory, the function works like `file.path()`, rooted at the working
directory:

``` r
setwd(pkg)
is_r_package
#> Root criterion: contains a file "DESCRIPTION" with contents matching "^Package: "
is_r_package$find_file()
#> [1] "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586"
is_r_package$find_file("tests", "testthat")
#> [1] "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586/tests/testthat"
```

This works identically when starting from a subdirectory:

``` r
setwd(file.path(pkg, "R"))
is_r_package$find_file()
#> [1] "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586"
is_r_package$find_file("tests", "testthat")
#> [1] "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586/tests/testthat"
```

There is one exception: if the first component passed to `find_file()`
is already an absolute path. This allows safely applying this function
to paths that may be absolute or relative:

``` r
setwd(file.path(pkg, "R"))
path <- is_r_package$find_file()
is_r_package$find_file(path, "tests", "testthat")
#> [1] "/private/var/folders/dj/yhk9rkx97wn_ykqtnmk18xvc0000gn/T/RtmpVBhMLR/file337b29a46586/tests/testthat"
```

As long as you are sure that your working directory is somewhere inside
your project, you can retrieve the project root.

## Installation and further reading

Install the package from CRAN:

``` r
install.package("rprojroot")
```

See the
[documentation](https://rprojroot.r-lib.org/articles/rprojroot.html) for
more detail.

------------------------------------------------------------------------

## Code of Conduct

Please note that the rprojroot project is released with a [Contributor
Code of Conduct](https://rprojroot.r-lib.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# [rprojroot](https://rprojroot.r-lib.org/)

<!-- badges: start -->

[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html) [![rcc](https://github.com/r-lib/rprojroot/workflows/rcc/badge.svg)](https://github.com/r-lib/rprojroot/actions) [![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/rprojroot)](https://cran.r-project.org/package=rprojroot) [![Codecov test coverage](https://codecov.io/gh/r-lib/rprojroot/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/rprojroot?branch=main)

<!-- badges: end -->

This package helps accessing files relative to a *project root* to [stop the working directory insanity](https://gist.github.com/jennybc/362f52446fe1ebc4c49f). It is a low-level helper package for the [here](https://here.r-lib.org/) package.

<pre class='chroma'>
<span class='kr'><a href='https://rdrr.io/r/base/library.html'>library</a></span><span class='o'>(</span><span class='nv'><a href='https://rprojroot.r-lib.org/'>rprojroot</a></span><span class='o'>)</span></pre>

## Example

The rprojroot package works best when you have a “project”: all related files contained in a subdirectory that can be categorized using a strict criterion. Let’s create a package for demonstration.

<pre class='chroma'>
<span class='nv'>dir</span> <span class='o'>&lt;-</span> <span class='nf'><a href='https://rdrr.io/r/base/tempfile.html'>tempfile</a></span><span class='o'>(</span><span class='o'>)</span>
<span class='nv'>pkg</span> <span class='o'>&lt;-</span> <span class='nf'>usethis</span><span class='nf'>::</span><span class='nf'><a href='https://usethis.r-lib.org/reference/create_package.html'>create_package</a></span><span class='o'>(</span><span class='nv'>dir</span><span class='o'>)</span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Creating </span><span style='color: #0000BB;'>'/tmp/RtmpBLE08t/file294c3c8acca7/'</span></span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Setting active project to </span><span style='color: #0000BB;'>'/tmp/RtmpBLE08t/file294c3c8acca7'</span></span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Creating </span><span style='color: #0000BB;'>'R/'</span></span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Writing </span><span style='color: #0000BB;'>'DESCRIPTION'</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Package</span><span>: file294c3c8acca7</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Title</span><span>: What the Package Does (One Line, Title Case)</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Version</span><span>: 0.0.0.9000</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Date</span><span>: 2020-11-08</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Authors@R</span><span> (parsed):</span></span>
<span class='c'>#&gt;     * Kirill Müller &lt;krlmlr+r@mailbox.org&gt; [aut, cre] (&lt;https://orcid.org/0000-0002-1416-3412&gt;)</span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Description</span><span>: What the package does (one paragraph).</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>License</span><span>: GPL-3</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>URL</span><span>: https://github.com/krlmlr/rprojroot,</span></span>
<span class='c'>#&gt;     https://krlmlr.github.io/rprojroot</span>
<span class='c'>#&gt; <span style='color: #0000BB;'>BugReports</span><span>: https://github.com/krlmlr/rprojroot/issues</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Encoding</span><span>: UTF-8</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>LazyData</span><span>: true</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>Roxygen</span><span>: list(markdown = TRUE)</span></span>
<span class='c'>#&gt; <span style='color: #0000BB;'>RoxygenNote</span><span>: 7.1.1.9000</span></span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Writing </span><span style='color: #0000BB;'>'NAMESPACE'</span></span>
<span class='c'>#&gt; <span style='color: #00BB00;'>✓</span><span> Setting active project to </span><span style='color: #0000BB;'>'&lt;no active project&gt;'</span></span></pre>

R packages satisfy the `is_r_package` criterion. A criterion is an object that contains a `find_file()` function. With `pkg` as working directory, the function works like [`file.path()`](https://rdrr.io/r/base/file.path.html), rooted at the working directory:

<pre class='chroma'>
<span class='nf'><a href='https://rdrr.io/r/base/getwd.html'>setwd</a></span><span class='o'>(</span><span class='nv'>pkg</span><span class='o'>)</span>
<span class='nv'>is_r_package</span>
<span class='c'>#&gt; Root criterion: contains a file `DESCRIPTION` with contents matching `^Package: `</span>
<span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "/tmp/RtmpBLE08t/file294c3c8acca7"</span>
<span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='s'>"tests"</span>, <span class='s'>"testthat"</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "/tmp/RtmpBLE08t/file294c3c8acca7/tests/testthat"</span></pre>

This works identically when starting from a subdirectory:

<pre class='chroma'>
<span class='nf'><a href='https://rdrr.io/r/base/getwd.html'>setwd</a></span><span class='o'>(</span><span class='nf'><a href='https://rdrr.io/r/base/file.path.html'>file.path</a></span><span class='o'>(</span><span class='nv'>pkg</span>, <span class='s'>"R"</span><span class='o'>)</span><span class='o'>)</span>
<span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "/tmp/RtmpBLE08t/file294c3c8acca7"</span>
<span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='s'>"tests"</span>, <span class='s'>"testthat"</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "/tmp/RtmpBLE08t/file294c3c8acca7/tests/testthat"</span></pre>

There is one exception: if the first component passed to `find_file()` is already an absolute path. This allows safely applying this function to paths that may be absolute or relative:

<pre class='chroma'>
<span class='nf'><a href='https://rdrr.io/r/base/getwd.html'>setwd</a></span><span class='o'>(</span><span class='nf'><a href='https://rdrr.io/r/base/file.path.html'>file.path</a></span><span class='o'>(</span><span class='nv'>pkg</span>, <span class='s'>"R"</span><span class='o'>)</span><span class='o'>)</span>
<span class='nv'>path</span> <span class='o'>&lt;-</span> <span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='o'>)</span>
<span class='nv'>is_r_package</span><span class='o'>$</span><span class='nf'>find_file</span><span class='o'>(</span><span class='nv'>path</span>, <span class='s'>"tests"</span>, <span class='s'>"testthat"</span><span class='o'>)</span>
<span class='c'>#&gt; [1] "/tmp/RtmpBLE08t/file294c3c8acca7/tests/testthat"</span></pre>

As long as you are sure that your working directory is somewhere inside your project, you can retrieve the project root.

## Installation and further reading

Install the package from CRAN:

<pre class='chroma'>
<span class='nf'>install.package</span><span class='o'>(</span><span class='s'>"rprojroot"</span><span class='o'>)</span></pre>

See the [documentation](https://rprojroot.r-lib.org/articles/rprojroot.html) for more detail.

------------------------------------------------------------------------

## Code of Conduct

Please note that the rprojroot project is released with a [Contributor Code of Conduct](https://rprojroot.r-lib.org/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms.

if (requireNamespace("rlang", quietly = TRUE)) {
  colon_colon <- `::`

  ensym <- rlang::ensym
  inject <- rlang::inject
  as_string <- rlang::as_string

  `::` <- function(x, y) {
    x_sym <- ensym(x)
    y_sym <- ensym(y)
    tryCatch(
      inject(colon_colon(!!x_sym, !!y_sym)),
      packageNotFoundError = function(e) {
        skip_if_not_installed(as_string(x_sym))
      }
    )
  }
}

#' File paths relative to the root of a directory hierarchy
#'
#' `find_root_file()` is a wrapper around [find_root()] that
#' appends an arbitrary number of path components to the root using
#' [base::file.path()].
#'
#' This function operates on the notion of relative paths.
#' The `...` argument is expected to contain a path relative to the root.
#' If the first path component passed to `...` is already an absolute path,
#' the `criterion` and `path` arguments are ignored,
#' and `...` is forwarded to [file.path()].
#'
#' @param criterion `[root_criterion]`\cr
#'   A criterion, one of the predefined [criteria]
#'   or created by [root_criterion()].
#'   Will be coerced using [as_root_criterion()].
#' @param path `[character(1)]`\cr
#'   The start directory.
#' @param ... `[character]`\cr
#'   Further path components passed to [file.path()].
#'   All arguments must be the same length or length one.
#' @return The normalized path of the root as specified by the search criteria,
#'   with the additional path components appended.
#'   Throws an error if no root is found.
#'
#' @examples
#' \dontrun{
#' find_package_root_file("tests", "testthat.R")
#' has_file("DESCRIPTION", "^Package: ")$find_file
#' has_file("DESCRIPTION", "^Package: ")$make_fix_file(".")
#' }
#'
#' @seealso [find_root()] [utils::glob2rx()] [base::file.path()]
#'
#' @export
find_root_file <- function(..., criterion, path = ".") {
  if (!missing(..1)) {
    abs <- is_absolute_path(..1)
    if (all(abs)) {
      return(path(...))
    }
    if (any(abs)) {
      stop("Combination of absolute and relative paths not supported.", call. = FALSE)
    }
  }

  root <- find_root(criterion = criterion, path = path)
  path(root, ...)
}

test_that("has_file", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(
        find_root_file("c", criterion = "b/a", path = path),
        file.path(hierarchy(2L), "c")
      )
      # Absolute paths are stripped
      expect_equal(
        find_root_file("/x", "y", criterion = "b/a", path = path),
        file.path("/x", "y")
      )
      expect_identical(
        find_root_file("c", NA, criterion = "b/a", path = path),
        NA_character_
      )
      expect_identical(
        find_root_file("c", character(), criterion = "b/a", path = path),
        character()
      )
      expect_error(
        find_root_file(letters[1:2], letters[1:3], criterion = "a", path = path)
      )
      expect_error(
        find_root_file(letters[1:2], character(), criterion = "a", path = path)
      )
      expect_error(
        find_root_file(c("b", "/x"), "c", criterion = "a", path = path),
        "absolute and relative"
      )
    }
  )
})

test_that("has_file", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(
    find_root_file("c", criterion = "b/a", path = path),
    file.path(hierarchy(2L), "c")
  )
  # Absolute paths are stripped
  expect_equal(
    find_root_file("/x", "y", criterion = "b/a", path = path),
    file.path("/x", "y")
  )
  expect_identical(
    find_root_file("c", NA, criterion = "b/a", path = path),
    NA_character_
  )
  expect_identical(
    find_root_file("c", character(), criterion = "b/a", path = path),
    character()
  )
  expect_error(
    find_root_file(letters[1:2], letters[1:3], criterion = "a", path = path)
  )
  expect_error(
    find_root_file(letters[1:2], character(), criterion = "a", path = path)
  )
  expect_error(
    find_root_file(c("b", "/x"), "c", criterion = "a", path = path),
    "absolute and relative"
  )
})

#' @details
#' See the "Value" section in [root_criterion()] for documentation
#' of root criterion objects, and  [criteria] for useful predefined
#' root criteria.
#'
#' @examples
#' criteria
#' \dontrun{
#' is_r_package$find_file("NAMESPACE")
#' root_fun <- is_r_package$make_fix_file()
#' root_fun("NAMESPACE")
#' }
"_PACKAGE"

list_files <- function(path, filename) {
  files <- dir(path = path, pattern = filename, all.files = TRUE, full.names = TRUE)
  dirs <- dir.exists(files)
  files <- files[!dirs]
  files
}

match_contents <- function(f, contents, n, fixed) {
  if (is.null(contents)) {
    return(TRUE)
  }

  fc <- readLines(f, n)
  any(grepl(contents, fc, fixed = fixed, useBytes = TRUE))
}

test_that("match_contents in non-native encoding", {
  desc <- c("Package: test", "Author: Kirill M\xFCller", "Encoding: latin1")
  Encoding(desc) <- "latin1"
  writeLines(desc, descfile <- tempfile(), useBytes = TRUE)
  expect_silent(expect_true(
    rprojroot:::match_contents(descfile,
      contents = "^Package: ",
      n = -1L, fixed = FALSE
    )
  ))
})

make_find_root_file <- function(criterion) {
  force(criterion)
  eval(bquote(function(..., path = ".") {
    find_root_file(..., criterion = criterion, path = path)
  }))
}

make_fix_root_file <- function(criterion, path, subdir = NULL) {
  root <- find_root(criterion = criterion, path = path)
  if (!is.null(subdir)) {
    root <- file.path(root, subdir)
  }
  eval(bquote(function(...) {
    if (!missing(..1)) {
      abs <- is_absolute_path(..1)
      if (all(abs)) {
        return(path(...))
      }
      if (any(abs)) {
        stop("Combination of absolute and relative paths not supported.", call. = FALSE)
      }
    }

    path(.(root), ...)
  }))
}

#' Is a directory the project root?
#'
#' Objects of the `root_criterion` class decide if a
#' given directory is a project root.
#'
#' Construct criteria using `root_criterion` in a very general fashion
#' by specifying a function with a `path` argument, and a description.
#'
#' @param testfun `[function|list(function)]`\cr
#'   A function with one parameter that returns `TRUE`
#'   if the directory specified by this parameter is the project root,
#'   and `FALSE` otherwise. Can also be a list of such functions.
#' @param desc `[character]`\cr
#'   A textual description of the test criterion, of the same length
#'   as `testfun`.
#' @param subdir `[character]`\cr
#'   If given, the criterion will also be tested in the subdirectories
#'   defined by this argument, in the order given.
#'   The first existing directory will be used as a starting point.
#'   This is used for the [is_testthat] criterion that needs to
#'   *descend* into `tests/testthat` if starting at the package root,
#'   but stay inside `tests/testthat` if called from a testthat test.
#'
#' @return
#' An S3 object of class `root_criterion` with the following members:
#'
#' @export
#'
#' @examples
#' root_criterion(function(path) file.exists(file.path(path, "somefile")), "has somefile")
#' has_file("DESCRIPTION")
#' is_r_package
#' \dontrun{
#' is_r_package$find_file
#' is_r_package$make_fix_file(".")
#' }
root_criterion <- function(testfun, desc, subdir = NULL) {
  testfun <- check_testfun(testfun)

  stopifnot(length(desc) == length(testfun))

  full_desc <- paste0(
    desc,
    if (!is.null(subdir)) {
      paste0(
        " (also look in subdirectories: ",
        paste0("`", subdir, "`", collapse = ", "),
        ")"
      )
    }
  )

  criterion <- structure(
    list(
      #' @return
      #' \describe{
      #'   \item{`testfun`}{The `testfun` argument}
      testfun = testfun,
      #'   \item{`desc`}{The `desc` argument}
      desc = full_desc,
      #'   \item{`subdir`}{The `subdir` argument}
      subdir = subdir
    ),
    class = "root_criterion"
  )

  #'   \item{`find_file`}{A function with `...` and `path` arguments
  #'     that returns a path relative to the root,
  #'     as specified by this criterion.
  #'     The optional `path` argument specifies the starting directory,
  #'     which defaults to `"."`.
  #'     The function forwards to [find_root_file()],
  #'     which passes `...` directly to `file.path()`
  #'     if the first argument is an absolute path.
  #'   }
  criterion$find_file <- make_find_root_file(criterion)
  #'   \item{`make_fix_file`}{A function with a `path` argument that
  #'      returns a function that finds paths relative to the root.  For a
  #'      criterion `cr`, the result of `cr$make_fix_file(".")(...)`
  #'      is identical to `cr$find_file(...)`. The function created by
  #'      `make_fix_file()` can be saved to a variable to be more independent
  #'      of the current working directory.
  #'   }
  #' }
  criterion$make_fix_file <-
    function(path = getwd(), subdir = NULL) {
      make_fix_root_file(criterion, path, subdir)
    }

  criterion
}

check_testfun <- function(testfun) {
  if (is.function(testfun)) {
    testfun <- list(testfun)
  }

  for (f in testfun) {
    if (!isTRUE(all.equal(names(formals(f)), "path"))) {
      stop("All functions in testfun must have exactly one argument 'path'", call. = FALSE)
    }
  }

  testfun
}

test_that("Shortcuts", {
  expect_equal(
    make_find_root_file("testthat.R")("testthat"),
    normalizePath(getwd(), winslash = "/")
  )

  R <- make_fix_root_file("testthat.R", getwd())

  oldwd <- withr::local_dir("~")

  expect_equal(
    normalizePath(R("testthat"), mustWork = TRUE),
    normalizePath(oldwd, mustWork = TRUE)
  )

  path <- R()
  expect_equal(
    normalizePath(R(path, "testthat"), mustWork = TRUE),
    normalizePath(oldwd, mustWork = TRUE)
  )
})

test_that("root_criterion", {
  expect_error(
    suppressWarnings(root_criterion(5, "Bogus")),
    "must have exactly one argument"
  )
  expect_error(root_criterion(identity, "Bogus"), "must have exactly one argument")
  expect_true(is_root_criterion(root_criterion(function(path) FALSE, "Never")))
})

test_that("Absolute paths are returned", {
  expect_equal(
    find_root("testthat.R"),
    normalizePath(find_root("testthat.R"), winslash = "/")
  )
})

#' Deprecated functions
#'
#' Use [as_root_criterion()] and [is_root_criterion()], respectively.
#'
#' @inheritParams as_root_criterion
#' @inheritParams is_root_criterion
#' @keywords internal
#' @name deprecated
#' @export
as.root_criterion <- function(...) {
  as_root_criterion(...)
}

#' @export
#' @rdname deprecated
is.root_criterion <- function(...) {
  is_root_criterion(...)
}

#' @rdname root_criterion
#' @param x `[object]`\cr
#'   An object.
#' @export
is_root_criterion <- function(x) {
  inherits(x, "root_criterion")
}

#' @rdname root_criterion
#' @export
as_root_criterion <- function(x) UseMethod("as_root_criterion", x)

#' @details
#' The `as_root_criterion()` function accepts objects of class
#' `root_criterion`, and character values; the latter will be
#' converted to criteria using `has_file`.
#'
#' @rdname root_criterion
#' @export
as_root_criterion.character <- function(x) {
  has_file(x)
}

#' @rdname root_criterion
#' @export
as_root_criterion.root_criterion <- identity

#' @export
as_root_criterion.default <- function(x) {
  stop("Cannot coerce ", x, " to type root_criterion.", call. = FALSE)
}

#' @export
format.root_criterion <- function(x, ...) {
  if (length(x$desc) > 1) {
    c("Root criterion: one of", paste0("- ", x$desc))
  } else {
    paste0("Root criterion: ", x$desc)
  }
}

#' @export
print.root_criterion <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' @export
#' @rdname root_criterion
#' @details Root criteria can be combined with the `|` operator. The result is a
#'   composite root criterion that requires either of the original criteria to
#'   match.
#' @param y `[object]`\cr
#'   An object.
`|.root_criterion` <- function(x, y) {
  stopifnot(is_root_criterion(y))

  root_criterion(
    c(x$testfun, y$testfun),
    c(x$desc, y$desc)
  )
}

#' Find the root of a directory hierarchy
#'
#' A \emph{root} is defined as a directory that contains a regular file
#' whose name matches a given pattern and which optionally contains a given text.
#' The search for a root starts at a given directory (the working directory
#' by default), and proceeds up the directory hierarchy.
#'
#' Starting from the working directory, the `find_root()` function searches
#' for the root.
#' If a root is found, the `...` arguments are used to construct a path;
#' thus, if no extra arguments are given, the root is returned.
#' If no root is found, an error is thrown.
#'
#' @inheritParams find_root_file
#' @return The normalized path of the root as specified by the search criterion.
#'   Throws an error if no root is found
#'
#' @examples
#' \dontrun{
#' find_root(has_file_pattern(
#'   pattern = glob2rx("DESCRIPTION"),
#'   contents = "^Package: "
#' ))
#' }
#'
#' @seealso [utils::glob2rx()] [file.path()]
#'
#' @export
find_root <- function(criterion, path = ".") {
  criterion <- as_root_criterion(criterion)

  start_path <- get_start_path(path, criterion$subdir)
  path <- start_path

  for (i in seq_len(.MAX_DEPTH)) {
    for (f in criterion$testfun) {
      if (f(path)) {
        return(path)
      }
    }

    if (is_fs_root(path)) {
      stop("No root directory found in ", start_path, " or its parent directories. ",
        paste(format(criterion), collapse = "\n"),
        call. = FALSE
      )
    }

    path <- dirname(path)
  }

  stop("Maximum search of ", .MAX_DEPTH, " exceeded. Last path: ", path, call. = FALSE)
}

.MAX_DEPTH <- 100L

get_start_path <- function(path, subdirs) {
  path <- normalizePath(path, winslash = "/", mustWork = TRUE)

  for (subdir in subdirs) {
    subdir_path <- file.path(path, subdir)
    if (dir.exists(subdir_path)) {
      return(subdir_path)
    }
  }

  path
}

# Borrowed from devtools
is_fs_root <- function(path) {
  identical(
    normalizePath(path, winslash = "/", mustWork = FALSE),
    normalizePath(dirname(path), winslash = "/", mustWork = FALSE)
  )
}

#' @rdname find_root
#' @description `get_root_desc()` returns the description of the criterion
#'   for a root path. This is especially useful for composite root criteria
#'   created with [|.root_criterion()].
#' @export
get_root_desc <- function(criterion, path) {
  for (i in seq_along(criterion$testfun)) {
    if (criterion$testfun[[i]](path)) {
      return(criterion$desc[[i]])
    }
  }

  stop("path is not a root. ",
    paste(format(criterion), collapse = "\n"),
    call. = FALSE
  )
}


format_lines <- function(n) {
  if (n == 1) "line" else paste0(n, " lines")
}

#' @details
#' The `has_file()` function constructs a criterion that checks for the
#' existence of a specific file (which itself can be in a subdirectory of the
#' root) with specific contents.
#'
#' @rdname root_criterion
#' @param filepath `[character(1)]`\cr
#'   File path (can contain directories).
#' @param contents,fixed `[character(1)]`\cr
#'   If `contents` is `NULL` (the default), file contents are not checked.
#'   Otherwise, `contents` is a regular expression
#'   (if `fixed` is `FALSE`) or a search string (if `fixed` is `TRUE`), and
#'   file contents are checked matching lines.
#' @param n `[integerish(1)]`\cr
#'   Maximum number of lines to read to check file contents.
#' @export
has_file <- function(filepath, contents = NULL, n = -1L, fixed = FALSE) {
  force(filepath)
  stopifnot(is.character(filepath), length(filepath) == 1)
  force(contents)
  if (!is.null(contents)) {
    stopifnot(is.character(contents), length(contents) == 1)
  }
  force(n)
  stopifnot(length(n) == 1)

  check_relative(filepath)

  testfun <- eval(bquote(function(path) {
    testfile <- file.path(path, .(filepath))
    if (!file.exists(testfile)) {
      return(FALSE)
    }
    if (dir.exists(testfile)) {
      return(FALSE)
    }
    match_contents(testfile, .(contents), .(n), .(fixed))
  }))

  desc <- paste0(
    "contains a file '", filepath, "'",
    if (!is.null(contents)) {
      paste0(
        " with contents ",
        if (!fixed) "matching ",
        "'", contents, "'",
        if (n >= 0L) paste0(" in the first ", format_lines(n))
      )
    }
  )

  root_criterion(testfun, desc)
}

#' @details
#' The `has_dir()` function constructs a criterion that checks for the
#' existence of a specific directory.
#'
#' @rdname root_criterion
#' @export
has_dir <- function(filepath) {
  force(filepath)
  stopifnot(is.character(filepath), length(filepath) == 1)

  check_relative(filepath)

  testfun <- eval(bquote(function(path) {
    testfile <- file.path(path, .(filepath))
    dir.exists(testfile)
  }))

  desc <- paste0("contains a directory '", filepath, "'")

  root_criterion(testfun, desc)
}

check_relative <- function(filepath) {
  if (is_absolute_path(filepath)) {
    stop("filepath must be a file or a relative path, not an absolute path.", call. = FALSE)
  }
}

#' @details
#' The `has_file_pattern()` function constructs a criterion that checks for the
#' existence of a file that matches a pattern, with specific contents.
#'
#' @rdname root_criterion
#' @param pattern `[character(1)]`\cr
#'   Regular expression to match the file name against.
#' @export
has_file_pattern <- function(pattern, contents = NULL, n = -1L, fixed = FALSE) {
  force(pattern)
  stopifnot(is.character(pattern), length(pattern) == 1)
  force(contents)
  if (!is.null(contents)) {
    stopifnot(is.character(contents), length(contents) == 1)
  }
  force(n)
  stopifnot(length(n) == 1)

  testfun <- eval(bquote(function(path) {
    files <- list_files(path, .(pattern))
    for (f in files) {
      if (!match_contents(f, .(contents), .(n), .(fixed))) {
        next
      }
      return(TRUE)
    }
    return(FALSE)
  }))

  desc <- paste0(
    "contains a file matching '", pattern, "'",
    if (!is.null(contents)) {
      paste0(
        " with contents ",
        if (!fixed) "matching ",
        "'", contents, "'",
        if (n >= 0L) paste0(" in the first ", format_lines(n))
      )
    }
  )

  root_criterion(testfun, desc)
}

#' @details
#' The `has_basename()` function constructs a criterion that checks if the
#' [base::basename()] of the root directory has a specific name,
#' with support for case-insensitive file systems.
#'
#' @rdname root_criterion
#' @param basename `[character(1)]`\cr
#'   The required name of the root directory.
#' @export
has_basename <- function(basename, subdir = NULL) {
  force(basename)

  testfun <- eval(bquote(function(path) {
    # Support case insensitive file systems.
    tolower(basename(path)) == tolower(.(basename)) && dir.exists(file.path(dirname(path), .(basename)))
  }))

  desc <- paste0('directory name is "', basename, '"')

  root_criterion(testfun, desc, subdir = subdir)
}

#' @export
is_rstudio_project <- has_file_pattern("[.]Rproj$", contents = "^Version: ", n = 1L)

#' @export
is_vscode_project <- has_file(".vscode/settings.json")

#' @export
is_r_package <- has_file("DESCRIPTION", contents = "^Package: ")

#' @export
is_remake_project <- has_file("remake.yml")

#' @export
is_drake_project <- has_dir(".drake")

#' @export
is_targets_project <- has_file("_targets.R")

#' @export
is_pkgdown_project <-
  has_file("_pkgdown.yml") |
    has_file("_pkgdown.yaml") |
    has_file("pkgdown/_pkgdown.yml") |
    has_file("pkgdown/_pkgdown.yaml") |
    has_file("inst/_pkgdown.yml") |
    has_file("inst/_pkgdown.yaml")

#' @export
is_renv_project <- has_file("renv.lock", contents = '"Packages":\\s*\\{')

#' @export
is_projectile_project <- has_file(".projectile")

#' @export
is_quarto_project <- has_file("_quarto.yml")

#' @export
is_git_root <- has_dir(".git") | has_file(".git", contents = "^gitdir: ")

#' @export
is_svn_root <- has_dir(".svn")

#' @export
is_vcs_root <- is_git_root | is_svn_root

#' @export
is_testthat <- has_basename("testthat", c("tests/testthat", "testthat"))

#' @export
from_wd <- root_criterion(function(path) TRUE, "from current working directory")

#' Prespecified criteria
#'
#' This is a collection of commonly used root criteria.
#'
#' @format NULL
#'
#' @export
criteria <- structure(
  list(
    is_rstudio_project = is_rstudio_project,
    is_vscode_project = is_vscode_project,
    is_r_package = is_r_package,
    is_remake_project = is_remake_project,
    is_pkgdown_project = is_pkgdown_project,
    is_renv_project = is_renv_project,
    is_projectile_project = is_projectile_project,
    is_quarto_project = is_quarto_project,
    is_git_root = is_git_root,
    is_svn_root = is_svn_root,
    is_vcs_root = is_vcs_root,
    is_testthat = is_testthat,
    from_wd = from_wd
  ),
  class = "root_criteria"
)

#' @export
#' @importFrom utils str
str.root_criteria <- function(object, ...) {
  str(lapply(object, format))
}

#' @details
#' `is_rstudio_project` looks for a file with extension `.Rproj`.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_rstudio_project"

#' @details
#' `is_vscode_project` looks for a `.vscode/settings.json` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_vscode_project"

#' @details
#' `is_r_package` looks for a `DESCRIPTION` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_r_package"

#' @details
#' `is_remake_project` looks for a `remake.yml` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_remake_project"

#' @details
#' `is_drake_project` looks for a `.drake` directory.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_drake_project"


#' @details
#' `is_targets_project` looks for a `_targets.R` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_targets_project"

#' @details
#' `is_pkgdown_project` looks for a `_pkgdown.yml`, `_pkgdown.yaml`, `pkgdown/_pkgdown.yml` and/or `inst/_pkgdown.yml` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_pkgdown_project"

#' @details
#' `is_renv_project` looks for an `renv.lock` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_renv_project"

#' @details
#' `is_projectile_project` looks for a `.projectile` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_projectile_project"

#' @details
#' `is_quarto_project` looks for a `_quarto.yml` file.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_quarto_project"

#' @details
#' `is_git_root` looks for a `.git` directory.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_git_root"

#' @details
#' `is_svn_root` looks for a `.svn` directory.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_svn_root"

#' @details
#' `is_vcs_root` looks for the root of a version control
#' system, currently only Git and SVN are supported.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_vcs_root"

#' @details
#' `is_testthat` looks for the `testthat` directory, works when
#'   developing, testing, and checking a package.
#'
#' @format NULL
#' @rdname criteria
#' @export
"is_testthat"

#' @details
#' `from_wd` uses the current working directory.
#'
#' @format NULL
#' @rdname criteria
#' @export
"from_wd"

test_that("is_root_criterion", {
  expect_true(is_root_criterion(has_file("DESCRIPTION")))
  expect_false(is_root_criterion("DESCRIPTION"))
  expect_true(is_root_criterion(as_root_criterion("DESCRIPTION")))
})

test_that("as_root_criterion", {
  reset_env <- function(x) {
    if (is.function(x)) {
      environment(x) <- .GlobalEnv
    } else if (is.list(x)) {
      x <- lapply(x, reset_env)
    }
    x
  }

  expect_equal(
    lapply(as_root_criterion("x"), reset_env),
    lapply(has_file("x"), reset_env)
  )
  expect_error(as_root_criterion(5), "Cannot coerce")
})

test_that("Formatting", {
  expect_snapshot(format(is_r_package))
  expect_snapshot(is_r_package)
  expect_snapshot(is_vcs_root)
  expect_snapshot(has_file("a", contents = "foo", fixed = TRUE))
  expect_snapshot(has_file_pattern("a.*b", contents = "foo", fixed = TRUE))
  expect_snapshot(criteria)
  expect_snapshot(str(criteria))
})

test_that("Combining criteria", {
  comb_crit <- is_r_package | is_rstudio_project

  expect_true(is_root_criterion(comb_crit))

  expect_snapshot(comb_crit)

  expect_equal(
    find_root(comb_crit, "hierarchy"),
    find_root(is_rstudio_project, "hierarchy/a")
  )
})

test_that("has_file", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root("a", path = path), hierarchy(3L))
      expect_equal(find_root("b", path = path), hierarchy(3L))
      expect_equal(find_root("b/a", path = path), hierarchy(2L))
      expect_equal(find_root("c", path = path), hierarchy(1L))
      expect_equal(find_root("d", path = path), hierarchy(4L))
      expect_equal(find_root(has_file("DESCRIPTION", "^Package: ", 1), path = path), hierarchy(1L))
      expect_equal(find_root(has_file("DESCRIPTION", "^Package: "), path = path), hierarchy(1L))
      expect_equal(find_root(has_file("DESCRIPTION", "package* does", fixed = TRUE), path = path), hierarchy(1L))
      expect_error(
        find_root("test-root.R", path = path),
        "No root directory found"
      )
      expect_error(
        find_root("rprojroot.Rproj", path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_file("e", "f"), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_file("e", "f", 1), path = path),
        "No root directory found"
      )
      expect_error(has_file("/a"), "absolute")
      TRUE
    }
  )
})

test_that("has_file_pattern", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(has_file_pattern(glob2rx("a")), path = path), hierarchy(3L))
      expect_equal(find_root(has_file_pattern(glob2rx("b")), path = path), hierarchy(3L))
      expect_equal(
        find_root(has_file_pattern("[ab]", "File b"), path = path),
        hierarchy(3L)
      )
      expect_equal(
        find_root(has_file_pattern("[ab]", "File b in root"), path = path),
        hierarchy(1L)
      )
      expect_equal(find_root(has_file_pattern(glob2rx("c")), path = path), hierarchy(1L))
      expect_equal(find_root(has_file_pattern(glob2rx("d")), path = path), hierarchy(4L))
      expect_equal(find_root(has_file_pattern(glob2rx("DES*ION"), "^Package: ", 1), path = path), hierarchy(1L))
      expect_equal(find_root(has_file_pattern(glob2rx("DESCRI?TION"), "^Package: "), path = path), hierarchy(1L))
      expect_equal(find_root(has_file_pattern(glob2rx("D?SCRIPTI?N"), "package* does", fixed = TRUE), path = path), hierarchy(1L))
      expect_error(
        find_root(has_file_pattern(glob2rx("test-root.R")), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_file_pattern(glob2rx("rprojroot.Rproj")), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_file_pattern(glob2rx("e"), "f"), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_file_pattern(glob2rx("e"), "f", 1), path = path),
        "No root directory found"
      )
    }
  )
})

test_that("has_dir", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(has_dir("a"), path = path), hierarchy(1L))
      expect_equal(find_root(has_dir("b"), path = path), hierarchy(2L))
      expect_equal(find_root(has_dir("c"), path = path), hierarchy(3L))
      expect_error(
        find_root(has_dir("e"), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_dir("rprojroot.Rproj"), path = path),
        "No root directory found"
      )
      expect_error(has_dir("/a"), "absolute")
    }
  )
})

test_that("has_basename", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(has_basename("a"), path = path), hierarchy(2L))
      expect_equal(find_root(has_basename("b"), path = path), hierarchy(3L))
      expect_equal(find_root(has_basename("c"), path = path), hierarchy(4L))
      expect_error(
        find_root(has_basename("d"), path = path),
        "No root directory found"
      )
      expect_error(
        find_root(has_basename("rprojroot.Rproj"), path = path),
        "No root directory found"
      )
    }
  )
})

test_that("concrete criteria", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  # HACK
  writeLines(character(), file.path(hierarchy(3L), ".projectile"))

  stop_path <- hierarchy(0L)
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(is_rstudio_project, path = path), hierarchy(1L))
      expect_equal(find_root(is_remake_project, path = path), hierarchy(2L))
      expect_equal(find_root(is_projectile_project, path = path), hierarchy(3L))
    }
  )
})

test_that("is_svn_root", {
  temp_dir <- tempfile("svn")
  unzip("vcs/svn.zip", exdir = temp_dir)
  wd <- normalizePath(temp_dir, winslash = "/")

  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "svn", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- normalizePath(tempdir(), winslash = "/")
  path <- hierarchy(4L)

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(is_svn_root, path = path), hierarchy(1L))
      expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
      expect_error(
        find_root(is_svn_root, path = hierarchy(0L)),
        "No root directory found"
      )
      expect_error(
        find_root(is_vcs_root, path = hierarchy(0L)),
        "No root directory found"
      )
    }
  )
})

setup_git_root <- function(separate_git_dir = FALSE) {
  temp_dir <- tempfile("git")
  unzip("vcs/git.zip", exdir = temp_dir)
  wd <- normalizePath(temp_dir, winslash = "/")

  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "git", "a", "b", "c")[seq_len(n + 1L)])
  }

  if (separate_git_dir) {
    # Copy .git dir to a separate location, then make a .git file.
    # (other_git_folder becomes a bare git repo)
    old_git_location <- file.path(wd, "git", ".git")
    new_git_location <- file.path(wd, "other_git_folder")
    file.rename(old_git_location, new_git_location)
    writeLines(paste("gitdir:", new_git_location), old_git_location)
  }
  return(hierarchy)
}

test_that("is_git_root", {
  hierarchy <- setup_git_root(separate_git_dir = FALSE)

  path <- hierarchy(4L)
  stop_path <- normalizePath(tempdir(), winslash = "/")

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(is_git_root, path = path), hierarchy(1L))
      expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
      expect_error(
        find_root(is_git_root, path = hierarchy(0L)),
        "No root directory found"
      )
      expect_error(
        find_root(is_vcs_root, path = hierarchy(0L)),
        "No root directory found"
      )
    }
  )
})

test_that("is_git_root for separated git directory", {
  hierarchy <- setup_git_root(separate_git_dir = TRUE)
  path <- hierarchy(4L)
  stop_path <- normalizePath(tempdir(), winslash = "/")

  mockr::with_mock(
    is_root = function(x) x == stop_path,
    {
      expect_equal(find_root(is_git_root, path = path), hierarchy(1L))
      expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
      expect_error(
        find_root(is_git_root, path = hierarchy(0L)),
        "No root directory found"
      )
      expect_error(
        find_root(is_vcs_root, path = hierarchy(0L)),
        "No root directory found"
      )
    }
  )
})

test_that("finds root", {
  skip_on_cran()
  # Checks that search for root actually terminates
  expect_error(
    find_root("9259cfa7884bf51eb9dd80b52c26dcdf9cd28e82"),
    "No root directory found"
  )
})

test_that("stops if depth reached", {
  find_root_mocked <- find_root
  mock_env <- new.env()
  mock_env$dirname <- identity
  environment(find_root_mocked) <- mock_env

  # Checks that search for root terminates for very deep hierarchies
  expect_error(find_root_mocked(""), "Maximum search of [0-9]+ exceeded")
})

test_that("is_root_criterion", {
  expect_true(is_root_criterion(has_file("DESCRIPTION")))
  expect_false(is_root_criterion("DESCRIPTION"))
  expect_true(is_root_criterion(as_root_criterion("DESCRIPTION")))
})

test_that("as_root_criterion", {
  reset_env <- function(x) {
    if (is.function(x)) {
      environment(x) <- .GlobalEnv
    } else if (is.list(x)) {
      x <- lapply(x, reset_env)
    }
    x
  }

  expect_equal(
    lapply(as_root_criterion("x"), reset_env),
    lapply(has_file("x"), reset_env)
  )
  expect_error(as_root_criterion(5), "Cannot coerce")
})

test_that("Formatting", {
  expect_snapshot(format(is_r_package))
  expect_snapshot(is_r_package)
  expect_snapshot(is_vcs_root)
  expect_snapshot(has_file("a", contents = "foo", fixed = TRUE))
  expect_snapshot(has_file_pattern("a.*b", contents = "foo", fixed = TRUE))
  expect_snapshot(criteria)
  expect_snapshot(str(criteria))
})

test_that("Combining criteria", {
  comb_crit <- is_r_package | is_rstudio_project

  expect_true(is_root_criterion(comb_crit))

  expect_snapshot(comb_crit)

  expect_equal(
    find_root(comb_crit, "hierarchy"),
    find_root(is_rstudio_project, "hierarchy/a")
  )
})

test_that("has_file", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root("a", path = path), hierarchy(3L))
  expect_equal(find_root("b", path = path), hierarchy(3L))
  expect_equal(find_root("b/a", path = path), hierarchy(2L))
  expect_equal(find_root("c", path = path), hierarchy(1L))
  expect_equal(find_root("d", path = path), hierarchy(4L))
  expect_equal(find_root(has_file("DESCRIPTION", "^Package: ", 1), path = path), hierarchy(1L))
  expect_equal(find_root(has_file("DESCRIPTION", "^Package: "), path = path), hierarchy(1L))
  expect_equal(find_root(has_file("DESCRIPTION", "package* does", fixed = TRUE), path = path), hierarchy(1L))
  expect_error(
    find_root("test-root.R", path = path),
    "No root directory found"
  )
  expect_error(
    find_root("rprojroot.Rproj", path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_file("e", "f"), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_file("e", "f", 1), path = path),
    "No root directory found"
  )
  expect_error(has_file("/a"), "absolute")
})

test_that("has_file_pattern", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(has_file_pattern(glob2rx("a")), path = path), hierarchy(3L))
  expect_equal(find_root(has_file_pattern(glob2rx("b")), path = path), hierarchy(3L))
  expect_equal(
    find_root(has_file_pattern("[ab]", "File b"), path = path),
    hierarchy(3L)
  )
  expect_equal(
    find_root(has_file_pattern("[ab]", "File b in root"), path = path),
    hierarchy(1L)
  )
  expect_equal(find_root(has_file_pattern(glob2rx("c")), path = path), hierarchy(1L))
  expect_equal(find_root(has_file_pattern(glob2rx("d")), path = path), hierarchy(4L))
  expect_equal(find_root(has_file_pattern(glob2rx("DES*ION"), "^Package: ", 1), path = path), hierarchy(1L))
  expect_equal(find_root(has_file_pattern(glob2rx("DESCRI?TION"), "^Package: "), path = path), hierarchy(1L))
  expect_equal(find_root(has_file_pattern(glob2rx("D?SCRIPTI?N"), "package* does", fixed = TRUE), path = path), hierarchy(1L))
  expect_error(
    find_root(has_file_pattern(glob2rx("test-root.R")), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_file_pattern(glob2rx("rprojroot.Rproj")), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_file_pattern(glob2rx("e"), "f"), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_file_pattern(glob2rx("e"), "f", 1), path = path),
    "No root directory found"
  )
})

test_that("has_dir", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(has_dir("a"), path = path), hierarchy(1L))
  expect_equal(find_root(has_dir("b"), path = path), hierarchy(2L))
  expect_equal(find_root(has_dir("c"), path = path), hierarchy(3L))
  expect_error(
    find_root(has_dir("e"), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_dir("rprojroot.Rproj"), path = path),
    "No root directory found"
  )
  expect_error(has_dir("/a"), "absolute")
})

test_that("has_basename", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- hierarchy(1L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(has_basename("a"), path = path), hierarchy(2L))
  expect_equal(find_root(has_basename("b"), path = path), hierarchy(3L))
  expect_equal(find_root(has_basename("c"), path = path), hierarchy(4L))
  expect_error(
    find_root(has_basename("d"), path = path),
    "No root directory found"
  )
  expect_error(
    find_root(has_basename("rprojroot.Rproj"), path = path),
    "No root directory found"
  )
})

test_that("concrete criteria", {
  wd <- normalizePath(getwd(), winslash = "/")
  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "hierarchy", "a", "b", "c")[seq_len(n + 1L)])
  }

  # HACK to prevent a NOTE like this from R CMD check:
  # checking for hidden files and directories ... NOTE
  # Found the following hidden files and directories:
  #   tests/testthat/hierarchy/.vscode
  dot_projectile <- file.path(hierarchy(3L), ".projectile")
  on.exit(file.remove(dot_projectile), add = TRUE)
  writeLines(character(), dot_projectile)

  dot_vscode <- file.path(hierarchy(1L), ".vscode")
  on.exit(unlink(dot_vscode, recursive = TRUE), add = TRUE)
  dir.create(dot_vscode)
  writeLines("{}", file.path(dot_vscode, "settings.json"))

  stop_path <- hierarchy(0L)
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(is_rstudio_project, path = path), hierarchy(1L))
  expect_equal(find_root(is_vscode_project, path = path), hierarchy(1L))
  expect_equal(find_root(is_remake_project, path = path), hierarchy(2L))
  expect_equal(find_root(is_projectile_project, path = path), hierarchy(3L))
})

test_that("is_svn_root", {
  temp_dir <- tempfile("svn")
  unzip("vcs/svn.zip", exdir = temp_dir)
  wd <- normalizePath(temp_dir, winslash = "/")

  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "svn", "a", "b", "c")[seq_len(n + 1L)])
  }

  stop_path <- normalizePath(tempdir(), winslash = "/")
  path <- hierarchy(4L)

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(is_svn_root, path = path), hierarchy(1L))
  expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
  expect_error(
    find_root(is_svn_root, path = hierarchy(0L)),
    "No root directory found"
  )
  expect_error(
    find_root(is_vcs_root, path = hierarchy(0L)),
    "No root directory found"
  )
})

setup_git_root <- function(separate_git_dir = FALSE) {
  temp_dir <- tempfile("git")
  unzip("vcs/git.zip", exdir = temp_dir)
  wd <- normalizePath(temp_dir, winslash = "/")

  hierarchy <- function(n = 0L) {
    do.call(file.path, list(wd, "git", "a", "b", "c")[seq_len(n + 1L)])
  }

  if (separate_git_dir) {
    # Copy .git dir to a separate location, then make a .git file.
    # (other_git_folder becomes a bare git repo)
    old_git_location <- file.path(wd, "git", ".git")
    new_git_location <- file.path(wd, "other_git_folder")
    file.rename(old_git_location, new_git_location)
    writeLines(paste("gitdir:", new_git_location), old_git_location)
  }
  return(hierarchy)
}

test_that("is_git_root", {
  hierarchy <- setup_git_root(separate_git_dir = FALSE)

  path <- hierarchy(4L)
  stop_path <- normalizePath(tempdir(), winslash = "/")

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(is_git_root, path = path), hierarchy(1L))
  expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
  expect_error(
    find_root(is_git_root, path = hierarchy(0L)),
    "No root directory found"
  )
  expect_error(
    find_root(is_vcs_root, path = hierarchy(0L)),
    "No root directory found"
  )
})

test_that("is_git_root for separated git directory", {
  hierarchy <- setup_git_root(separate_git_dir = TRUE)
  path <- hierarchy(4L)
  stop_path <- normalizePath(tempdir(), winslash = "/")

  local_mocked_bindings(is_fs_root = function(x) x == stop_path)

  expect_equal(find_root(is_git_root, path = path), hierarchy(1L))
  expect_equal(find_root(is_vcs_root, path = path), hierarchy(1L))
  expect_error(
    find_root(is_git_root, path = hierarchy(0L)),
    "No root directory found"
  )
  expect_error(
    find_root(is_vcs_root, path = hierarchy(0L)),
    "No root directory found"
  )
})

test_that("finds root", {
  skip_on_cran()
  # Checks that search for root actually terminates
  expect_error(
    find_root("9259cfa7884bf51eb9dd80b52c26dcdf9cd28e82"),
    "No root directory found"
  )
})

test_that("stops if depth reached", {
  find_root_mocked <- find_root
  mock_env <- new.env()
  mock_env$dirname <- identity
  environment(find_root_mocked) <- mock_env

  # Checks that search for root terminates for very deep hierarchies
  expect_error(find_root_mocked(""), "Maximum search of [0-9]+ exceeded")
})

# The exact style of quoting here (double inside single as opposed to nested
# double quotes) makes generating the here docs *much less fiddly*.
# See https://github.com/r-lib/here/pull/130
test_that("is_renv_project prints regex-friendly-ish text", {
  expect_snapshot(is_renv_project)
})

# Modeled after fs::path()
path <- function(...) {
  dots <- list(...)

  if (!is.null(names(dots)) && any(names(dots) != "")) {
    warning("Arguments must be unnamed", call. = FALSE)
  }

  # Different recycling rules for zero-length vectors
  lengths <- lengths(dots)
  if (any(lengths == 0) && all(lengths %in% 0:1)) {
    return(character())
  }

  # Side effect: check recycling rules
  component_df <- as.data.frame(dots, stringsAsFactors = FALSE)

  missing <- apply(is.na(component_df), 1, any)

  components <- lapply(component_df, function(x) enc2utf8(as.character(x)))

  out <- do.call(file.path, components)
  out[missing] <- NA_character_
  Encoding(out) <- "UTF-8"
  out
}

# Adapted from fs
describe("path", {
  it("returns paths UTF-8 encoded", {
    skip_on_os("solaris")
    expect_equal(Encoding(path("föö")), "UTF-8")
  })

  it("returns paths UTF-8 encoded 2", {
    skip_on_os("solaris")
    skip_on_os("windows")
    expect_equal(Encoding(path("\U4F60\U597D.R")), "UTF-8")
  })

  it("returns empty strings for empty inputs", {
    expect_equal(path(""), "")
    expect_equal(path(character()), character())
    expect_equal(path("foo", character(), "bar"), character())
  })

  it("propagates NA strings", {
    expect_equal(path(NA_character_), NA_character_)
    expect_equal(path("foo", NA_character_), NA_character_)
    expect_equal(path(c("foo", "bar"), c("baz", NA_character_)), c("foo/baz", NA_character_))
  })

  it("does not double paths", {
    expect_equal(path("", "foo"), "/foo")

    # This could be a UNC path, so we keep the doubled path.
    expect_equal(path("//foo", "bar"), "//foo/bar")
  })

  it("errors on paths which are too long", {
    expect_error(path(paste(rep("a", 100000), collapse = "")))
  })

  it("follows recycling rules", {
    expect_equal(path("foo", character()), character())
    expect_equal(path("foo", "bar"), "foo/bar")
    expect_equal(path("foo", c("bar", "baz")), c("foo/bar", "foo/baz"))
    expect_equal(path(c("foo", "qux"), c("bar", "baz")), c("foo/bar", "qux/baz"))

    expect_error(path(c("foo", "qux", "foo2"), c("bar", "baz")))
  })
})

# nocov start
#' Determines the path of the currently running script
#'
#' @description
#' `r lifecycle::badge("soft-deprecated")`
#'
#' \R does not store nor export the path of the currently running
#'   script.  This is an attempt to circumvent this limitation by applying
#'   heuristics (such as call stack and argument inspection) that work in many
#'   cases.
#' **CAVEAT**: Use this function only if your workflow does not permit other
#' solution: if a script needs to know its location, it should be set outside
#' the context of the script if possible.
#'
#' @details This functions currently work only if the script was `source`d,
#'   processed with `knitr`,
#'   or run with `Rscript` or using the `--file` parameter to the
#'   `R` executable.  For code run with `Rscript`, the exact value
#'   of the parameter passed to `Rscript` is returned.
#'
#' @section Life cycle:
#'
#' These functions are now available in the \pkg{whereami} package.
#'
#' @return The path of the currently running script, NULL if it cannot be
#'   determined.
#' @seealso [base::source()], [utils::Rscript()], [base::getwd()]
#' @references [https://stackoverflow.com/q/1815606/946850]()
#' @author Kirill Müller, Hadley Wickham, Michael R. Head
#' @keywords internal
#' @examples
#' \dontrun{
#' thisfile()
#' }
#' @export
thisfile <- function() {
  lifecycle::deprecate_soft(
    "2.0.0",
    "rprojroot::thisfile()",
    "whereami::thisfile()"
  )

  if (!is.null(res <- thisfile_source())) {
    res
  } else if (!is.null(res <- thisfile_r())) {
    res
  } else if (!is.null(res <- thisfile_rscript())) {
    res
  } else if (!is.null(res <- thisfile_knit())) {
    res
  } else {
    NULL
  }
}

#' @rdname thisfile
#' @export
thisfile_source <- function() {
  lifecycle::deprecate_soft(
    "2.0.0",
    "rprojroot::thisfile_source()",
    "whereami::thisfile_source()"
  )

  for (i in -(1:sys.nframe())) {
    if (identical(args(sys.function(i)), args(base::source))) {
      return(normalizePath(sys.frame(i)$ofile))
    }
  }

  NULL
}

#' @rdname thisfile
#' @importFrom utils tail
#' @export
thisfile_r <- function() {
  lifecycle::deprecate_soft(
    "2.0.0",
    "rprojroot::thisfile_r()",
    "whereami::thisfile_r()"
  )

  cmd_args <- commandArgs(trailingOnly = FALSE)
  if (!grepl("^R(?:|term)(?:|[.]exe)$", basename(cmd_args[[1L]]), ignore.case = TRUE)) {
    return(NULL)
  }

  cmd_args_trailing <- commandArgs(trailingOnly = TRUE)
  leading_idx <-
    seq.int(from = 1, length.out = length(cmd_args) - length(cmd_args_trailing))
  cmd_args <- cmd_args[leading_idx]
  file_idx <- c(which(cmd_args == "-f") + 1, which(grepl("^--file=", cmd_args)))
  res <- gsub("^(?:|--file=)(.*)$", "\\1", cmd_args[file_idx])

  # If multiple --file arguments are given, R uses the last one
  res <- tail(res[res != ""], 1)
  if (length(res) > 0) {
    return(res)
  }

  NULL
}

#' @rdname thisfile
#' @importFrom utils tail
#' @export
thisfile_rscript <- function() {
  lifecycle::deprecate_soft(
    "2.0.0",
    "rprojroot::thisfile_rscript()",
    "whereami::thisfile_rscript()"
  )

  cmd_args <- commandArgs(trailingOnly = FALSE)
  if (!grepl("^R(?:term|script)(?:|[.]exe)$", basename(cmd_args[[1L]]), ignore.case = TRUE)) {
    return(NULL)
  }

  cmd_args_trailing <- commandArgs(trailingOnly = TRUE)
  leading_idx <-
    seq.int(from = 1, length.out = length(cmd_args) - length(cmd_args_trailing))
  cmd_args <- cmd_args[leading_idx]
  res <- gsub("^(?:--file=(.*)|.*)$", "\\1", cmd_args)

  # If multiple --file arguments are given, R uses the last one
  res <- tail(res[res != ""], 1)
  if (length(res) > 0) {
    return(res)
  }

  NULL
}

#' @rdname thisfile
#' @export
thisfile_knit <- function() {
  lifecycle::deprecate_soft(
    "2.0.0",
    "rprojroot::thisfile_knit()",
    "whereami::thisfile_knit()"
  )

  if (requireNamespace("knitr")) {
    return(knitr::current_input(dir = TRUE))
  }

  NULL
}
# nocov end

#' @rdname find_root_file
#' @export
find_rstudio_root_file <- is_rstudio_project$find_file

#' @rdname find_root_file
#' @export
find_package_root_file <- is_r_package$find_file

#' @rdname find_root_file
#' @export
find_remake_root_file <- is_remake_project$find_file

#' @rdname find_root_file
#' @export
find_testthat_root_file <- is_testthat$find_file

test_that("is_testthat", {
  expect_snapshot(is_testthat)

  testthat_path <- normalizePath("package/tests/testthat", winslash = "/")
  expect_equal(is_testthat$find_file(path = "package"), testthat_path)
  expect_equal(is_testthat$find_file(path = "package/tests"), testthat_path)
  expect_equal(is_testthat$find_file(path = "package/tests/testthat"), testthat_path)
})

test_that("dogfood", {
  expect_true(file.exists(is_testthat$find_file("hierarchy", "a", "b", "c", "d")))
})

# From fs
describe("is_absolute_path", {
  it("detects windows absolute paths", {
    expect_true(is_absolute_path("c:\\"))
    expect_true(is_absolute_path("c:/"))
    expect_true(is_absolute_path("P:/"))
    expect_true(is_absolute_path("P:\\"))
    expect_true(is_absolute_path("\\\\server\\mountpoint\\"))
    expect_true(is_absolute_path("\\foo"))
    expect_true(is_absolute_path("\\foo\\bar"))
  })
  it("detects posix absolute paths", {
    expect_false(is_absolute_path(""))
    expect_false(is_absolute_path("foo/bar"))
    expect_false(is_absolute_path("./foo/bar"))
    expect_false(is_absolute_path("../foo/bar"))

    expect_true(is_absolute_path("/"))
    expect_true(is_absolute_path("/foo"))
    expect_true(is_absolute_path("/foo/bar"))
    expect_true(is_absolute_path("~/foo/bar"))
  })
})

# From fs
describe("is_absolute_path", {
  it("detects windows absolute paths", {
    expect_true(is_absolute_path("c:\\"))
    expect_true(is_absolute_path("c:/"))
    expect_true(is_absolute_path("P:/"))
    expect_true(is_absolute_path("P:\\"))
    expect_true(is_absolute_path("\\\\server\\mountpoint\\"))
    expect_true(is_absolute_path("\\foo"))
    expect_true(is_absolute_path("\\foo\\bar"))
  })
  it("detects posix absolute paths", {
    expect_false(is_absolute_path(""))
    expect_false(is_absolute_path("foo/bar"))
    expect_false(is_absolute_path("./foo/bar"))
    expect_false(is_absolute_path("../foo/bar"))

    expect_true(is_absolute_path("/"))
    expect_true(is_absolute_path("/foo"))
    expect_true(is_absolute_path("/foo/bar"))
    expect_true(is_absolute_path("~/foo/bar"))
  })
})

---
title: "Finding files in project subdirectories"
author: "Kirill Müller"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Finding files in project subdirectories}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The *rprojroot* package solves a seemingly trivial but annoying problem
that occurs sooner or later
in any largish project:
How to find files in subdirectories?
Ideally, file paths are relative to the *project root*.

Unfortunately, we cannot always be sure about the current working directory:
For instance, in RStudio it's sometimes:

- the project root (when running R scripts),
- a subdirectory (when building vignettes),
- again the project root (when executing chunks of a vignette).

```{r}
basename(getwd())
```

In some cases, it's even outside the project root.

This vignette starts with a very brief summary that helps you get started,
followed by a longer description of the features.

## TL;DR

What is your project: An R package?

```{r}
rprojroot::is_r_package
```

Or an RStudio project?

```{r}
rprojroot::is_rstudio_project
```

Or something else?

```{r}
rprojroot::has_file(".git/index")
```

For now, we assume it's an R package:

```{r}
root <- rprojroot::is_r_package
```

The `root` object contains a function that helps locating files below the root
of your package, regardless of your current working directory.
If you are sure that your working directory is somewhere below your project's root,
use the `root$find_file()` function.
In this example here, we're starting in the `vignettes` subdirectory and find the original `DESCRIPTION` file:

```{r}
basename(getwd())
readLines(root$find_file("DESCRIPTION"), 3)
```

There is one exception: if the first component passed to `find_file()` is already an absolute path.
This allows safely applying this function to paths that may be absolute or relative:

```{r}
path <- root$find_file()
readLines(root$find_file(path, "DESCRIPTION"), 3)
```

You can also
construct an accessor to your root using the `root$make_fix_file()` function:

```{r}
root_file <- root$make_fix_file()
```


Note that `root_file()` is a *function* that works just like `$find_file()` but
will find the files even if the current working directory is outside your project:

```{r}
withr::with_dir(
  "../..",
  readLines(root_file("DESCRIPTION"), 3)
)
```

If you know the absolute path of some directory below your project,
but cannot be sure of your current working directory,
pass that absolute path to `root$make_fix_file()`:

```r
root_file <- root$make_fix_file("C:\\Users\\User Name\\...")
```

As a last resort, you can get the path of standalone R scripts or vignettes
using the `thisfile()` function:

```r
root_file <- root$make_fix_file(dirname(thisfile()))
```

The remainder of this vignette describes implementation details and advanced features.


## Project root

We assume a self-contained project
where all files and directories are located below a common *root* directory.
Also, there should be a way to unambiguously identify this root directory.
(Often, the root contains a regular file whose name matches a given pattern,
and/or whose contents match another pattern.)
In this case, the following method reliably finds our project root:

- Start the search in any subdirectory of our project
- Proceed up the directory hierarchy until the root directory has been identified

The Git version control system (and probably many other tools) use a similar
approach: A Git command can be executed from within any subdirectory of a
repository.


### A simple example

The `find_root()` function implements the core functionality.
It returns the path to the first directory that matches the filtering criteria,
or throws an error if there is no such directory.
Filtering criteria are constructed in a generic fashion using the 
`root_criterion()` function,
the `has_file()` function constructs a criterion that checks for the presence
of a file with a specific name and specific contents.

```{r}
library(rprojroot)

# List all files and directories below the root
dir(find_root(has_file("DESCRIPTION")))
```

#### Relative paths to a stable root

Here we illustrate the power of *rprojroot* by demonstrating how to access the same file from two different working directories. Let your project be a package called `pkgname` and consider the desired file `rrmake.R` at `pkgname/R/rrmake.R`. First, we show how to access from the `vignettes` directory, and then from the `tests/testthat` directory.


##### Example A: From `vignettes`

When your working directory is `pkgname/vignettes`, you can access the `rrmake.R` file by:

1. Supplying a pathname relative to your working directory. Here's two ways to do that:

```{r, eval = FALSE}
rel_path_from_vignettes <- "../R/rrmake.R"
rel_path_from_vignettes <- file.path("..", "R", "rrmake.R") ## identical
```

2. Supplying a pathname to the file relative from the root of the package, e.g.,

```{r, eval = FALSE}
rel_path_from_root <- "R/rrmake.R"
rel_path_from_root <- file.path("R", "rrmake.R") ## identical
```

This second method requires finding the root of the package, which can be done with the `has_file()` function:

```{r}
has_file("DESCRIPTION")
```

So, using *rprojroot* you can specify the path relative from root in the following manner:

```{r}
# Specify a path/to/file relative to the root
rel_path_from_root <- find_root_file("R", "rrmake.R", criterion = has_file("DESCRIPTION"))
```


##### Example B: From `tests/testthat`

When your working directory is `pkgname/tests/testthat`, you can access the `rrmake.R` file by:

1. Supplying a pathname relative to your working directory.

```{r, eval = FALSE}
rel_path_from_testthat <- "../../R/rrmake.R"
```

Note that this is different than in the previous example! However, the second method is the same...

2. Supplying a pathname to the file relative from the root of the package. With *rprojroot*, this is the exact same as in the previous example.


```{r}
# Specify a path/to/file relative to the root
rel_path_from_root <- find_root_file("R", "rrmake.R", criterion = has_file("DESCRIPTION"))
```

##### Summary of Examples A and B

Since Examples A and B used different working directories, `rel_path_from_vignettes` and `rel_path_from_testthat` were different. This is an issue when trying to re-use the same code. This issue is solved by using *rprojroot*: the function `find_root_file()` finds a file relative from the root, where the root is determined from checking the criterion with `has_file()`. 


Note that the follow code produces identical results when building the vignette *and* when sourcing the chunk in RStudio, provided that the current working directory is the project root or anywhere below. So, we can check to make sure that *rprojroot* has successfully determined the correct path:

```{r}
# Specify a path/to/file relative to the root
rel_path_from_root <- find_root_file("R", "rrmake.R", criterion = has_file("DESCRIPTION"))

# Find a file relative to the root
file.exists(rel_path_from_root)
```


### Criteria

The `has_file()` function (and the more general `root_criterion()`)
both return an S3 object of class `root_criterion`:

```{r}
has_file("DESCRIPTION")
```

In addition, character values are coerced to `has_file` criteria by default, this coercion is applied automatically by `find_root()`.
(This feature is used by the introductory example.)

```{r}
as_root_criterion("DESCRIPTION")
```

The return value of these functions can be stored and reused;
in fact, the package provides `r length(criteria)` such criteria:

```{r}
criteria
```

Defining new criteria is easy:

```{r}
has_license <- has_file("LICENSE")
has_license

is_projecttemplate_project <- has_file("config/global.dcf", "^version: ")
is_projecttemplate_project
```

You can also combine criteria via the `|` operator:

```{r}
is_r_package | is_rstudio_project
```



### Shortcuts

To avoid specifying the search criteria for the project root every time,
shortcut functions can be created.
The `find_package_root_file()` is a shortcut for
`find_root_file(..., criterion = is_r_package)`:

```{r}
# Print first lines of the source for this document
head(readLines(find_package_root_file("vignettes", "rprojroot.Rmd")))
```

To save typing effort, define a shorter alias:

```{r}
P <- find_package_root_file

# Use a shorter alias
file.exists(P("vignettes", "rprojroot.Rmd"))
```

Each criterion actually contains a function that allows finding a file below the root specified by this criterion.
As our project does not have a file named `LICENSE`, querying the root results in an error:

```{r error = TRUE}
# Use the has_license criterion to find the root
R <- has_license$find_file
R

# Our package does not have a LICENSE file, trying to find the root results in an error
R()
```


### Fixed root

We can also create a function
that computes a path relative to the root *at creation time*.

```{r eval = (Sys.getenv("IN_PKGDOWN") != "")}
# Define a function that computes file paths below the current root
F <- is_r_package$make_fix_file()
F

# Show contents of the NAMESPACE file in our project
readLines(F("NAMESPACE"))
```

This is a more robust alternative to `$find_file()`, because it *fixes* the project
directory when `$make_fix_file()` is called, instead of searching for it every
time.  (For that reason it is also slightly faster, but I doubt this matters
in practice.)

This function can be used even if we later change the working directory to somewhere outside the project:

```{r eval = (Sys.getenv("IN_PKGDOWN") != "")}
# Print the size of the namespace file, working directory outside the project
withr::with_dir(
  "../..",
  file.size(F("NAMESPACE"))
)
```

The `make_fix_file()` member function also accepts an optional `path` argument,
in case you know your project's root but the current working directory is somewhere outside.
The path to the current script or `knitr` document can be obtained using the `thisfile()` function, but it's much easier and much more robust to just run your scripts with the working directory somewhere below your project root.


## `testthat` files

Tests run with [`testthat`](https://cran.r-project.org/package=testthat)
commonly use files that live below the `tests/testthat` directory.
Ideally, this should work in the following situation:

- During package development (working directory: package root)
- When testing with `devtools::test()` (working directory: `tests/testthat`)
- When running `R CMD check` (working directory: a renamed recursive copy of `tests`)

The `is_testthat` criterion allows robust lookup of test files.

```{r}
is_testthat
```

The example code below lists all files in the
[hierarchy](https://github.com/r-lib/rprojroot/tree/main/tests/testthat/hierarchy)
test directory.
It uses two project root lookups in total,
so that it also works when rendering the vignette (*sigh*):

```{r}
dir(is_testthat$find_file("hierarchy", path = is_r_package$find_file()))
```

### Another example: custom testing utilities

The hassle of using saved data files for testing is made even easier by using *rprojroot* in a utility function. For example, suppose you have a testing file at `tests/testthat/test_my_fun.R` which tests the `my_fun()` function:

```{r, eval = FALSE}
my_fun_run <- do.call(my_fun, my_args)

testthat::test_that(
  "my_fun() returns expected output",
  testthat::expect_equal(
    my_fun_run,
    expected_output
  )
)
```

There are two pieces of information that you'll need every time `test_my_fun.R` is run: `my_args` and `expected_output`. Typically, these objects are saved to `.Rdata` files and saved to the same subdirectory. For example, you could save them to `my_args.Rdata` and `expected_output.Rdata` under the `tests/testthat/testing_data` subdirectory. And, to find them easily in any contexts, you can use *rprojroot*!

Since all of the data files live in the same subdirectory, you can create a utility function `get_my_path()` that will always look in that directory for these types of files. And, since the *testthat* package will look for and source the `tests/testthat/helper.R` file before running any tests, you can place a `get_my_path()` in this file and use it throughout your tests:

```{r, eval = FALSE}
## saved to tests/testthat/helper.R
get_my_path <- function(file_name) {
  rprojroot::find_testthat_root_file(
    "testing_data", filename
  )
}
```

Now you can ask `get_my_path()` to find your important data files by using the function within your test scripts!

```{r, eval = FALSE}
## Find the correct path with your custom rprojroot helper function
path_to_my_args_file <- get_my_path("my_args.Rdata")

## Load the input arguments
load(file = path_to_my_args_file)

## Run the function with those arguments
my_fun_run <- do.call(my_fun, my_args)

## Load the historical expectation with the helper
load(file = get_my_path("expected_output.Rdata"))

## Pass all tests and achieve nirvana
testthat::test_that(
  "my_fun() returns expected output",
  testthat::expect_equal(
    my_fun_run,
    expected_output
  )
)
```

For an example in the wild, see the [`test_sheet()` function](https://github.com/tidyverse/readxl/blob/0d9ad4f570f6580ff716e0e9ba5048447048e9f0/tests/testthat/helper.R#L1-L3) in the *readxl* package.

## Summary

The *rprojroot* package allows easy access to files below a project root
if the project root can be identified easily, e.g. if it is the only directory
in the whole hierarchy that contains a specific file.
This is a robust solution for finding files in largish projects
with a subdirectory hierarchy if the current working directory cannot be assumed
fixed.
(However, at least initially, the current working directory must be
somewhere below the project root.)


## Acknowledgement

This package was inspired by the gist
["Stop the working directory insanity"](https://gist.github.com/jennybc/362f52446fe1ebc4c49f)
by Jennifer Bryan, and by the way Git knows where its files are.

# Formatting

    Code
      format(is_r_package)
    Output
      [1] "Root criterion: contains a file \"DESCRIPTION\" with contents matching \"^Package: \""

---

    Code
      is_r_package
    Output
      Root criterion: contains a file "DESCRIPTION" with contents matching "^Package: "

---

    Code
      is_vcs_root
    Output
      Root criterion: one of
      - contains a directory ".git"
      - contains a file ".git" with contents matching "^gitdir: "
      - contains a directory ".svn"

---

    Code
      has_file("a", contents = "foo", fixed = TRUE)
    Output
      Root criterion: contains a file "a" with contents "foo"

---

    Code
      has_file_pattern("a.*b", contents = "foo", fixed = TRUE)
    Output
      Root criterion: contains a file matching "a.*b" with contents "foo"

---

    Code
      criteria
    Output
      $is_rstudio_project
      Root criterion: contains a file matching "[.]Rproj$" with contents matching "^Version: " in the first line
      
      $is_r_package
      Root criterion: contains a file "DESCRIPTION" with contents matching "^Package: "
      
      $is_remake_project
      Root criterion: contains a file "remake.yml"
      
      $is_pkgdown_project
      Root criterion: one of
      - contains a file "_pkgdown.yml"
      - contains a file "_pkgdown.yaml"
      - contains a file "pkgdown/_pkgdown.yml"
      - contains a file "inst/_pkgdown.yml"
      
      $is_renv_project
      Root criterion: contains a file "renv.lock" with contents matching ""Packages":\s*\{"
      
      $is_projectile_project
      Root criterion: contains a file ".projectile"
      
      $is_quarto_project
      Root criterion: contains a file "_quarto.yml"
      
      $is_git_root
      Root criterion: one of
      - contains a directory ".git"
      - contains a file ".git" with contents matching "^gitdir: "
      
      $is_svn_root
      Root criterion: contains a directory ".svn"
      
      $is_vcs_root
      Root criterion: one of
      - contains a directory ".git"
      - contains a file ".git" with contents matching "^gitdir: "
      - contains a directory ".svn"
      
      $is_testthat
      Root criterion: directory name is "testthat" (also look in subdirectories: `tests/testthat`, `testthat`)
      
      $from_wd
      Root criterion: from current working directory
      
      attr(,"class")
      [1] "root_criteria"

---

    Code
      str(criteria)
    Output
      List of 12
       $ is_rstudio_project   : chr "Root criterion: contains a file matching \"[.]Rproj$\" with contents matching \"^Version: \" in the first line"
       $ is_r_package         : chr "Root criterion: contains a file \"DESCRIPTION\" with contents matching \"^Package: \""
       $ is_remake_project    : chr "Root criterion: contains a file \"remake.yml\""
       $ is_pkgdown_project   : chr [1:5] "Root criterion: one of" "- contains a file \"_pkgdown.yml\"" "- contains a file \"_pkgdown.yaml\"" "- contains a file \"pkgdown/_pkgdown.yml\"" ...
       $ is_renv_project      : chr "Root criterion: contains a file \"renv.lock\" with contents matching \"\"Packages\":\\s*\\{\""
       $ is_projectile_project: chr "Root criterion: contains a file \".projectile\""
       $ is_quarto_project    : chr "Root criterion: contains a file \"_quarto.yml\""
       $ is_git_root          : chr [1:3] "Root criterion: one of" "- contains a directory \".git\"" "- contains a file \".git\" with contents matching \"^gitdir: \""
       $ is_svn_root          : chr "Root criterion: contains a directory \".svn\""
       $ is_vcs_root          : chr [1:4] "Root criterion: one of" "- contains a directory \".git\"" "- contains a file \".git\" with contents matching \"^gitdir: \"" "- contains a directory \".svn\""
       $ is_testthat          : chr "Root criterion: directory name is \"testthat\" (also look in subdirectories: `tests/testthat`, `testthat`)"
       $ from_wd              : chr "Root criterion: from current working directory"

# Combining criteria

    Code
      comb_crit
    Output
      Root criterion: one of
      - contains a file "DESCRIPTION" with contents matching "^Package: "
      - contains a file matching "[.]Rproj$" with contents matching "^Version: " in the first line

# Formatting

    Code
      format(is_r_package)
    Output
      [1] "Root criterion: contains a file 'DESCRIPTION' with contents matching '^Package: '"

---

    Code
      is_r_package
    Output
      Root criterion: contains a file 'DESCRIPTION' with contents matching '^Package: '

---

    Code
      is_vcs_root
    Output
      Root criterion: one of
      - contains a directory '.git'
      - contains a file '.git' with contents matching '^gitdir: '
      - contains a directory '.svn'

---

    Code
      has_file("a", contents = "foo", fixed = TRUE)
    Output
      Root criterion: contains a file 'a' with contents 'foo'

---

    Code
      has_file_pattern("a.*b", contents = "foo", fixed = TRUE)
    Output
      Root criterion: contains a file matching 'a.*b' with contents 'foo'

---

    Code
      criteria
    Output
      $is_rstudio_project
      Root criterion: contains a file matching '[.]Rproj$' with contents matching '^Version: ' in the first line
      
      $is_vscode_project
      Root criterion: contains a file '.vscode/settings.json'
      
      $is_r_package
      Root criterion: contains a file 'DESCRIPTION' with contents matching '^Package: '
      
      $is_remake_project
      Root criterion: contains a file 'remake.yml'
      
      $is_pkgdown_project
      Root criterion: one of
      - contains a file '_pkgdown.yml'
      - contains a file '_pkgdown.yaml'
      - contains a file 'pkgdown/_pkgdown.yml'
      - contains a file 'pkgdown/_pkgdown.yaml'
      - contains a file 'inst/_pkgdown.yml'
      - contains a file 'inst/_pkgdown.yaml'
      
      $is_renv_project
      Root criterion: contains a file 'renv.lock' with contents matching '"Packages":\s*\{'
      
      $is_projectile_project
      Root criterion: contains a file '.projectile'
      
      $is_quarto_project
      Root criterion: contains a file '_quarto.yml'
      
      $is_git_root
      Root criterion: one of
      - contains a directory '.git'
      - contains a file '.git' with contents matching '^gitdir: '
      
      $is_svn_root
      Root criterion: contains a directory '.svn'
      
      $is_vcs_root
      Root criterion: one of
      - contains a directory '.git'
      - contains a file '.git' with contents matching '^gitdir: '
      - contains a directory '.svn'
      
      $is_testthat
      Root criterion: directory name is "testthat" (also look in subdirectories: `tests/testthat`, `testthat`)
      
      $from_wd
      Root criterion: from current working directory
      
      attr(,"class")
      [1] "root_criteria"

---

    Code
      str(criteria)
    Output
      List of 13
       $ is_rstudio_project   : chr "Root criterion: contains a file matching '[.]Rproj$' with contents matching '^Version: ' in the first line"
       $ is_vscode_project    : chr "Root criterion: contains a file '.vscode/settings.json'"
       $ is_r_package         : chr "Root criterion: contains a file 'DESCRIPTION' with contents matching '^Package: '"
       $ is_remake_project    : chr "Root criterion: contains a file 'remake.yml'"
       $ is_pkgdown_project   : chr [1:7] "Root criterion: one of" "- contains a file '_pkgdown.yml'" "- contains a file '_pkgdown.yaml'" "- contains a file 'pkgdown/_pkgdown.yml'" ...
       $ is_renv_project      : chr "Root criterion: contains a file 'renv.lock' with contents matching '\"Packages\":\\s*\\{'"
       $ is_projectile_project: chr "Root criterion: contains a file '.projectile'"
       $ is_quarto_project    : chr "Root criterion: contains a file '_quarto.yml'"
       $ is_git_root          : chr [1:3] "Root criterion: one of" "- contains a directory '.git'" "- contains a file '.git' with contents matching '^gitdir: '"
       $ is_svn_root          : chr "Root criterion: contains a directory '.svn'"
       $ is_vcs_root          : chr [1:4] "Root criterion: one of" "- contains a directory '.git'" "- contains a file '.git' with contents matching '^gitdir: '" "- contains a directory '.svn'"
       $ is_testthat          : chr "Root criterion: directory name is \"testthat\" (also look in subdirectories: `tests/testthat`, `testthat`)"
       $ from_wd              : chr "Root criterion: from current working directory"

# Combining criteria

    Code
      comb_crit
    Output
      Root criterion: one of
      - contains a file 'DESCRIPTION' with contents matching '^Package: '
      - contains a file matching '[.]Rproj$' with contents matching '^Version: ' in the first line

# is_renv_project prints regex-friendly-ish text

    Code
      is_renv_project
    Output
      Root criterion: contains a file 'renv.lock' with contents matching '"Packages":\s*\{'

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\name{find_root}
\alias{find_root}
\alias{get_root_desc}
\title{Find the root of a directory hierarchy}
\usage{
find_root(criterion, path = ".")

get_root_desc(criterion, path)
}
\arguments{
\item{criterion}{\verb{[root_criterion]}\cr
A criterion, one of the predefined \link{criteria}
or created by \code{\link[=root_criterion]{root_criterion()}}.
Will be coerced using \code{\link[=as_root_criterion]{as_root_criterion()}}.}

\item{path}{\verb{[character(1)]}\cr
The start directory.}
}
\value{
The normalized path of the root as specified by the search criterion.
Throws an error if no root is found
}
\description{
A \emph{root} is defined as a directory that contains a regular file
whose name matches a given pattern and which optionally contains a given text.
The search for a root starts at a given directory (the working directory
by default), and proceeds up the directory hierarchy.

\code{get_root_desc()} returns the description of the criterion
for a root path. This is especially useful for composite root criteria
created with \code{\link[=|.root_criterion]{|.root_criterion()}}.
}
\details{
Starting from the working directory, the \code{find_root()} function searches
for the root.
If a root is found, the \code{...} arguments are used to construct a path;
thus, if no extra arguments are given, the root is returned.
If no root is found, an error is thrown.
}
\examples{
\dontrun{
find_root(has_file_pattern(
  pattern = glob2rx("DESCRIPTION"),
  contents = "^Package: "
))
}

}
\seealso{
\code{\link[utils:glob2rx]{utils::glob2rx()}} \code{\link[=file.path]{file.path()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rprojroot-package.R
\docType{package}
\name{rprojroot-package}
\alias{rprojroot}
\alias{rprojroot-package}
\title{rprojroot: Finding Files in Project Subdirectories}
\description{
Robust, reliable and flexible paths to files below a project root. The 'root' of a project is defined as a directory that matches a certain criterion, e.g., it contains a certain regular file.
}
\details{
See the "Value" section in \code{\link[=root_criterion]{root_criterion()}} for documentation
of root criterion objects, and  \link{criteria} for useful predefined
root criteria.
}
\examples{
criteria
\dontrun{
is_r_package$find_file("NAMESPACE")
root_fun <- is_r_package$make_fix_file()
root_fun("NAMESPACE")
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rprojroot.r-lib.org/}
  \item \url{https://github.com/r-lib/rprojroot}
  \item Report bugs at \url{https://github.com/r-lib/rprojroot/issues}
}

}
\author{
\strong{Maintainer}: Kirill Müller \email{kirill@cynkra.com} (\href{https://orcid.org/0000-0002-1416-3412}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thisfile.R
\name{thisfile}
\alias{thisfile}
\alias{thisfile_source}
\alias{thisfile_r}
\alias{thisfile_rscript}
\alias{thisfile_knit}
\title{Determines the path of the currently running script}
\usage{
thisfile()

thisfile_source()

thisfile_r()

thisfile_rscript()

thisfile_knit()
}
\value{
The path of the currently running script, NULL if it cannot be
determined.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#soft-deprecated}{\figure{lifecycle-soft-deprecated.svg}{options: alt='[Soft-deprecated]'}}}{\strong{[Soft-deprecated]}}

\R does not store nor export the path of the currently running
script.  This is an attempt to circumvent this limitation by applying
heuristics (such as call stack and argument inspection) that work in many
cases.
\strong{CAVEAT}: Use this function only if your workflow does not permit other
solution: if a script needs to know its location, it should be set outside
the context of the script if possible.
}
\details{
This functions currently work only if the script was \code{source}d,
processed with \code{knitr},
or run with \code{Rscript} or using the \code{--file} parameter to the
\code{R} executable.  For code run with \code{Rscript}, the exact value
of the parameter passed to \code{Rscript} is returned.
}
\section{Life cycle}{


These functions are now available in the \pkg{whereami} package.
}

\examples{
\dontrun{
thisfile()
}
}
\references{
\url{https://stackoverflow.com/q/1815606/946850}
}
\seealso{
\code{\link[base:source]{base::source()}}, \code{\link[utils:Rscript]{utils::Rscript()}}, \code{\link[base:getwd]{base::getwd()}}
}
\author{
Kirill Müller, Hadley Wickham, Michael R. Head
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root.R
\docType{data}
\name{criteria}
\alias{criteria}
\alias{is_rstudio_project}
\alias{is_vscode_project}
\alias{is_r_package}
\alias{is_remake_project}
\alias{is_drake_project}
\alias{is_targets_project}
\alias{is_pkgdown_project}
\alias{is_renv_project}
\alias{is_projectile_project}
\alias{is_quarto_project}
\alias{is_git_root}
\alias{is_svn_root}
\alias{is_vcs_root}
\alias{is_testthat}
\alias{from_wd}
\title{Prespecified criteria}
\usage{
criteria

is_rstudio_project

is_vscode_project

is_r_package

is_remake_project

is_drake_project

is_targets_project

is_pkgdown_project

is_renv_project

is_projectile_project

is_quarto_project

is_git_root

is_svn_root

is_vcs_root

is_testthat

from_wd
}
\description{
This is a collection of commonly used root criteria.
}
\details{
\code{is_rstudio_project} looks for a file with extension \code{.Rproj}.

\code{is_vscode_project} looks for a \code{.vscode/settings.json} file.

\code{is_r_package} looks for a \code{DESCRIPTION} file.

\code{is_remake_project} looks for a \code{remake.yml} file.

\code{is_drake_project} looks for a \code{.drake} directory.

\code{is_targets_project} looks for a \verb{_targets.R} file.

\code{is_pkgdown_project} looks for a \verb{_pkgdown.yml}, \verb{_pkgdown.yaml}, \verb{pkgdown/_pkgdown.yml} and/or \verb{inst/_pkgdown.yml} file.

\code{is_renv_project} looks for an \code{renv.lock} file.

\code{is_projectile_project} looks for a \code{.projectile} file.

\code{is_quarto_project} looks for a \verb{_quarto.yml} file.

\code{is_git_root} looks for a \code{.git} directory.

\code{is_svn_root} looks for a \code{.svn} directory.

\code{is_vcs_root} looks for the root of a version control
system, currently only Git and SVN are supported.

\code{is_testthat} looks for the \code{testthat} directory, works when
developing, testing, and checking a package.

\code{from_wd} uses the current working directory.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{as.root_criterion}
\alias{is.root_criterion}
\title{Deprecated functions}
\usage{
as.root_criterion(...)

is.root_criterion(...)
}
\description{
Use \code{\link[=as_root_criterion]{as_root_criterion()}} and \code{\link[=is_root_criterion]{is_root_criterion()}}, respectively.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/criterion.R, R/root.R
\name{root_criterion}
\alias{root_criterion}
\alias{is_root_criterion}
\alias{as_root_criterion}
\alias{as_root_criterion.character}
\alias{as_root_criterion.root_criterion}
\alias{|.root_criterion}
\alias{has_file}
\alias{has_dir}
\alias{has_file_pattern}
\alias{has_basename}
\title{Is a directory the project root?}
\usage{
root_criterion(testfun, desc, subdir = NULL)

is_root_criterion(x)

as_root_criterion(x)

\method{as_root_criterion}{character}(x)

\method{as_root_criterion}{root_criterion}(x)

\method{|}{root_criterion}(x, y)

has_file(filepath, contents = NULL, n = -1L, fixed = FALSE)

has_dir(filepath)

has_file_pattern(pattern, contents = NULL, n = -1L, fixed = FALSE)

has_basename(basename, subdir = NULL)
}
\arguments{
\item{testfun}{\verb{[function|list(function)]}\cr
A function with one parameter that returns \code{TRUE}
if the directory specified by this parameter is the project root,
and \code{FALSE} otherwise. Can also be a list of such functions.}

\item{desc}{\verb{[character]}\cr
A textual description of the test criterion, of the same length
as \code{testfun}.}

\item{subdir}{\verb{[character]}\cr
If given, the criterion will also be tested in the subdirectories
defined by this argument, in the order given.
The first existing directory will be used as a starting point.
This is used for the \link{is_testthat} criterion that needs to
\emph{descend} into \code{tests/testthat} if starting at the package root,
but stay inside \code{tests/testthat} if called from a testthat test.}

\item{x}{\verb{[object]}\cr
An object.}

\item{y}{\verb{[object]}\cr
An object.}

\item{filepath}{\verb{[character(1)]}\cr
File path (can contain directories).}

\item{contents, fixed}{\verb{[character(1)]}\cr
If \code{contents} is \code{NULL} (the default), file contents are not checked.
Otherwise, \code{contents} is a regular expression
(if \code{fixed} is \code{FALSE}) or a search string (if \code{fixed} is \code{TRUE}), and
file contents are checked matching lines.}

\item{n}{\verb{[integerish(1)]}\cr
Maximum number of lines to read to check file contents.}

\item{pattern}{\verb{[character(1)]}\cr
Regular expression to match the file name against.}

\item{basename}{\verb{[character(1)]}\cr
The required name of the root directory.}
}
\value{
An S3 object of class \code{root_criterion} with the following members:

\describe{
\item{\code{testfun}}{The \code{testfun} argument}
\item{\code{desc}}{The \code{desc} argument}
\item{\code{subdir}}{The \code{subdir} argument}
\item{\code{find_file}}{A function with \code{...} and \code{path} arguments
that returns a path relative to the root,
as specified by this criterion.
The optional \code{path} argument specifies the starting directory,
which defaults to \code{"."}.
The function forwards to \code{\link[=find_root_file]{find_root_file()}},
which passes \code{...} directly to \code{file.path()}
if the first argument is an absolute path.
}
\item{\code{make_fix_file}}{A function with a \code{path} argument that
returns a function that finds paths relative to the root.  For a
criterion \code{cr}, the result of \code{cr$make_fix_file(".")(...)}
is identical to \code{cr$find_file(...)}. The function created by
\code{make_fix_file()} can be saved to a variable to be more independent
of the current working directory.
}
}
}
\description{
Objects of the \code{root_criterion} class decide if a
given directory is a project root.
}
\details{
Construct criteria using \code{root_criterion} in a very general fashion
by specifying a function with a \code{path} argument, and a description.

The \code{as_root_criterion()} function accepts objects of class
\code{root_criterion}, and character values; the latter will be
converted to criteria using \code{has_file}.

Root criteria can be combined with the \code{|} operator. The result is a
composite root criterion that requires either of the original criteria to
match.

The \code{has_file()} function constructs a criterion that checks for the
existence of a specific file (which itself can be in a subdirectory of the
root) with specific contents.

The \code{has_dir()} function constructs a criterion that checks for the
existence of a specific directory.

The \code{has_file_pattern()} function constructs a criterion that checks for the
existence of a file that matches a pattern, with specific contents.

The \code{has_basename()} function constructs a criterion that checks if the
\code{\link[base:basename]{base::basename()}} of the root directory has a specific name,
with support for case-insensitive file systems.
}
\examples{
root_criterion(function(path) file.exists(file.path(path, "somefile")), "has somefile")
has_file("DESCRIPTION")
is_r_package
\dontrun{
is_r_package$find_file
is_r_package$make_fix_file(".")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R, R/shortcut.R
\name{find_root_file}
\alias{find_root_file}
\alias{find_rstudio_root_file}
\alias{find_package_root_file}
\alias{find_remake_root_file}
\alias{find_testthat_root_file}
\title{File paths relative to the root of a directory hierarchy}
\usage{
find_root_file(..., criterion, path = ".")

find_rstudio_root_file(..., path = ".")

find_package_root_file(..., path = ".")

find_remake_root_file(..., path = ".")

find_testthat_root_file(..., path = ".")
}
\arguments{
\item{...}{\verb{[character]}\cr
Further path components passed to \code{\link[=file.path]{file.path()}}.
All arguments must be the same length or length one.}

\item{criterion}{\verb{[root_criterion]}\cr
A criterion, one of the predefined \link{criteria}
or created by \code{\link[=root_criterion]{root_criterion()}}.
Will be coerced using \code{\link[=as_root_criterion]{as_root_criterion()}}.}

\item{path}{\verb{[character(1)]}\cr
The start directory.}
}
\value{
The normalized path of the root as specified by the search criteria,
with the additional path components appended.
Throws an error if no root is found.
}
\description{
\code{find_root_file()} is a wrapper around \code{\link[=find_root]{find_root()}} that
appends an arbitrary number of path components to the root using
\code{\link[base:file.path]{base::file.path()}}.
}
\details{
This function operates on the notion of relative paths.
The \code{...} argument is expected to contain a path relative to the root.
If the first path component passed to \code{...} is already an absolute path,
the \code{criterion} and \code{path} arguments are ignored,
and \code{...} is forwarded to \code{\link[=file.path]{file.path()}}.
}
\examples{
\dontrun{
find_package_root_file("tests", "testthat.R")
has_file("DESCRIPTION", "^Package: ")$find_file
has_file("DESCRIPTION", "^Package: ")$make_fix_file(".")
}

}
\seealso{
\code{\link[=find_root]{find_root()}} \code{\link[utils:glob2rx]{utils::glob2rx()}} \code{\link[base:file.path]{base::file.path()}}
}
